#ifndef HUFFMANNCODEC_H
#define HUFFMANNCODEC_H

/*
 * REFLEX - Real-time Event FLow EXecutive
 *
 * A lightweight operating system for deeply embedded systems.
 *
 *
 * Class(ses): HuffmanCodec
 * Revision: $Id:$
 * Author: Steffen Schulze
 *
 * Description: Huffmancoding implementation
 *
 * (c) Steffen Schulze, BTU Cottbus 2005
 */


#include "compress/CompressorCodec.h"

/**
 * Implementation of the Burrows-Wheeler-Transformation.
 * First count each charactervalue in inputdata and then creates
 * with this knowledge a binary tree using the huffman-algorithm.
 * by using this tree it encode the inputdata to new bitfield and
 * set the coded tree before the outputmessage.
 * When starting decoding it translate the coded tree back into the
 * original binary tree and decode the bitfield with its help back
 * to the uncompressed data.
 *
 * Decodierung und Encodierung ueber einen Huffmanbaum.
 */
class HuffmanCodec : public CompressorCodec
{

	public:
		HuffmanCodec()
		{
			#ifdef TEXTOUTPUT
			textOutputMode = 1;
			#endif
		}
		/**
		 * Encode-method
		 *
		 * Implements the virtual method encode of the baseclass CompressionCodec
		 */
		virtual	UINTMAX encode(BYTE* inputBuffer, UINTMAX inputBufferSize, BYTE* outputBuffer, UINTMAX outputBufferSize);
		
		/**
		 * Decode-method
		 *
		 * Implements the virtual method decode of the baseclass CompressionCodec
		 */
		virtual	UINTMAX decode(BYTE* inputBuffer, UINTMAX inputBufferSize, BYTE* outputBuffer, UINTMAX outputBufferSize);
		
	protected:
		/**
		 * Struct that represent the bitcode for a character.
		 */
		struct CharCode
		{
			/*
			 * the length of the bitcode
			 */
			UINT16	bitcount;
			/*
			 * the bitcode beginning rightside
			 */
			UINTMAX bitcode;
			/*
			 * index of the character
			 */
			UINT16	index;
		};

		
		
		/**
		 * Struct to represent a node of a binary tree in a flat
		 * array
		 */
		struct CharNode
		{
			/*
			 * point to the index of the childs. 0 means there is no child
			 */
			UINT16 child[2];
		};
		
		
		/**
		 * Node in a tree used by finding the bitcode for the charactervalues
		 */
		struct CharCodeNode
		{
			
			/*
			 * the current bitcode beginning rightside
			 */
			CharCode	code;
			
			/*
			 * count of the current charactervalue
			 */
			UINTMAX		count;
			
			/*
			 * index to the next node in this branch
			 */
			UINT16		buddy;
		};
		
		
		/**
		 * Creates a huffmantree by using the counts of the charactervalues
		 */
		void	findHuffmanTree(UINT16 charCounter[], BYTE* charIndex, UINT16 charIndexCount, CharCode* charCodes);
			/**
			 * Search in a branch the node that is topmost in the field
			 */
			UINT16	findTopBuddy(UINT16 start, CharCodeNode* charCodeNodes);
			/**
			 * Merge 2 branches to one
			 */
			void	mergeBranches(UINT16 branch1, UINT16 branch2, CharCodeNode* charCodeNodes, UINT16 charIndexCount);
				/**
				 * Fill the nodes of a branch with the new data
				 */
				void	tagBranch(UINT16 branch, UINTMAX count, CharCodeNode* charCodeNodes);
						
			/**
			 * Sort a tree by the deep of his nodes
			 */
			void	sortTree(CharCodeNode* charCodeNodes, UINT16 charIndexCount);
		
		/**
		 * Create the bitcode using the the deep of the nodes.
		 */
		void	buildHuffmanTree(CharCode* charCodes, UINT16 charIndexCount);

	

	#ifdef TEXTDEBUG
	private:
		//////////////////////////
		// Debug
		void	printCharCodes(CharCode* charCodes, UINT16 charCodeCount);
		void	printCharCodeNodes(CharCodeNode* charCodeNodes, UINT16 charIndexCount);
		// Debug
		//////////////////////////
		
	protected:
		int textOutputMode;
	
	#endif

};
#endif
